'''memory.py                31 March 2006'''

import os, random, math, itertools
from perstypes import Episode, Link, Pred, Obj
random.seed(os.urandom(99))
#from persparms import pi_nr, theta_f

def calculate_similarity(hebbs, driver):
    '''THIS FUNCTION IS DEPRECATED.

    Calculate MIPs.  Because the target units are the same as perception, the
    mapped units should have cos(t,s) = 1.  This will only be calculated over
    mapped units, so N(t) == N(s).'''

    # cos(t,s) = dot(x,y) / (len(x)*len(y))

    cosines = []

    for hebb in hebbs.itervalues():
        if driver in hebb.owners and 'world' in hebb.owners:
            if type(hebb.owners[driver]) is Pred or type(hebb.owners[driver]) is Obj:
                if hebb.perm_wt > 0.8:
                    target_op = set(hebb.owners[driver].semantic.keys())
                    world_op = set(hebb.owners['world'].semantic.keys())

                    # Get intersection of, and symmetric difference between, OPs
                    intersect = target_op & world_op
                    diff = target_op ^ world_op

                    # Build vectors of 1s and 0s                
                    target_v = [1 for i in intersect]
                    world_v = [1 for i in intersect]
                    for i in diff:
                        target_v.append(1)
                        world_v.append(0)

                    # Calculate cosine
                    dot = sum(x*y for x,y in itertools.izip(target_v, world_v))
                    len_t = math.sqrt(sum(pow(i,2) for i in target_v))
                    len_w = math.sqrt(sum(pow(i,2) for i in world_v))
                    cosines.append(dot/(len_t * len_w))

    if any(i for i in cosines if i < 0.99):
        consistent_mapping = False
    else:
        consistent_mapping = True

    return consistent_mapping



def reinforce_causal_power(analogs, driver, beta_cp):
    '''THIS FUNCTION IS DEPRECATED.

    This function is a modified verison of the hebb activation
    function.  beta_cp fills the role of hebb.act and the
    learning rate in that function.  It can be +/-, allowing the
    learning of positive and negative causal power.'''
    
    for group in analogs[driver].groups:
        if type(group) is Link:
            delta_cp = (1.1 - abs(group.causal_power)) * beta_cp
            cp = group.causal_power + delta_cp
            if -1.0 <= cp <= 1.0:
                group.causal_power = cp
            elif cp > 1.0:
                group.causal_power = 1.0
            else:
                group.causal_power = -1.0

    if 'schema' in analogs:
        for group in analogs['schema'].groups:
            if group.reinforcer is not None:
                if type(group) is Link:
                    delta_cp = (1.1 - abs(group.causal_power)) * beta_cp
                    cp = group.causal_power + delta_cp
                    if -1.0 <= cp <= 1.0:
                        group.causal_power = cp
                    elif cp > 1.0:
                        group.causal_power = 1.0
                    else:
                        group.causal_power = -1.0
                        
    for group in analogs['ltm'].groups:
        if group.reinforcer == True and type(group) is Link:
            delta_cp = (1.1 - abs(group.causal_power)) * beta_cp
            cp = group.causal_power + delta_cp
            if -1.0 <= cp <= 1.0:
                group.causal_power = cp
            elif cp > 1.0:
                group.causal_power = 1.0
            else:
                group.causal_power = -1.0
            

def reinforce_action(analogs, hebbs, beta, action_success_flag=True, goal_consistent_mapping=False,
                     penalty=0.0):
    '''THIS FUNCTION IS DEPRECATED.

    This function is called in two differenct situations.  The first is when an
    inferred action can't happen at all, which should cause an immediate penalty
    for the rule.  The second is when the inferred effect is mapped to
    perception.  Depending on what happens, the rule may be positively or
    negatively reinforced.

    Value propogates back one Episode, from effect to cause.

    Penalty is a small negative value used to prompt quick convergence
    on the most efficient path.'''

    if action_success_flag == False:
        # This is not currently being invoked
        for group in analogs['target'].groups:
            if any(i for i in group.semantic.iterkeys() if i.name == 'cause'):
                reward = -1.0
                error = reward - group.value
                group.value += beta * error
                
        for group in analogs['ltm'].groups:
            if group.reinforcer == True:
                if any(i for i in group.semantic.iterkeys() if i.name == 'cause'):
                    reward = -1.0
                    error = reward - group.value
                    group.value += beta * error
    else:
        # Get reward if we are at the goal
        reward = penalty
        new_state_value = 0.0
        if goal_consistent_mapping == True:
            for group in analogs['goal'].groups:
                if any(i for i in group.semantic.iterkeys() if i.name == 'effect'):
                    reward = group.reward

        # Get value of effect's equivalent cause if it exists
        for group in analogs['ltm'].groups:
            if group.reinforcer == True:
                if any(i for i in group.semantic.iterkeys() if i.name == 'effect'):
                    if group.equivalence is not None:
                        for episode in group.equivalence.episode:
                            if any(i for i in group.semantic.iterkeys() if i.name == 'cause'):
                                new_state_value = episode.value

        # This will fail silently if the inference was done from source to perception
        for group in analogs['target'].groups:
            if any(i for i in group.semantic.iterkeys() if i.name == 'cause'):
                error = reward + new_state_value - group.value
                group.value += beta * error
                'updated current state value:', group.value

        for group in analogs['ltm'].groups:
            if group.reinforcer == True:
                if any(i for i in group.semantic.iterkeys() if i.name == 'cause'):
                    error = reward + new_state_value - group.value
                    group.value += beta * error
                    'updated LTM state value:', group.value

    #error = r_new_state + v_new_state - v_old_state
    #v_old_state = v_old_state + beta * error


def update_prop_max_act(analogs):
    for prop in analogs['ltm'].props:
        prop.get_max_act()

def update_group_act_memory(analogs):
    for group in analogs['ltm'].groups:
        group.update_act_memory()

def get_retrieval_fit(analogs, direction):
    '''Want to get similarity str at link level or at cause group level'''
    for group in analogs['ltm'].groups:
        if type(group) is Episode:
            group.get_similarity_strength(direction)


def retrieve_analog(analogs, driver, pi_nr):
    '''This function retrieves groups probabilistically as a function of their
    similarity to the driver and their value.  Because values can be +/-, and
    are potentially unbounded in both directions, values are run through the
    Gibbs softmax version of the Luce choice rule before being combined with
    similarities.  The final probability of retrieval is based on an arithmetic
    Luce choice of similarity X Gibbs value.  This process gives a probability
    that scales well with all combinations of similarity and value.

    Retrieval probability scales linearly in both Similarity Strength and Gibbs
    value.  The inputs themselves are nonlinear; similarity strength is the product
    of a sigmoid function, and Gibbs value is exponetial.'''

    # Get groups and their similarity strengths; currently, only cause groups trigger
    # retrieval
    # Retrieve whole sequence by best single-group match
        
    groups = []
    similarity = []
    gibbs_value = []
    for group in analogs['ltm'].groups:
        if type(group) is Episode and group.inhibited == False:
            #if any(i.name for i in group.semantic.iterkeys() if i.name == 'cause'):
            groups.append(group)
            similarity.append(group.similarity_strength)
            gibbs_value.append(math.exp(group.value))

    '''If values are bounded [-1:1], then the Gibbs softmax calculation reduces
    to (e^value) for each value.  If values are allowed unbounded range, the
    full Gibbs calculation is e^value/sum(e^values).  The code to implement the
    full version is:
    values = [i.value for i in analogs['ltm'].groups.values()]
    gibbs_value = [math.exp(i)/sum(math.exp(j) for j in values) for i in values]'''
    
    # Arithmetic Luce choice on product of similarity strength and Gibbs value
    s_x_gv = [x*y for x,y in itertools.izip(similarity,gibbs_value)]
    #p_retrieve = [i/sum(s_x_gv) for i in s_x_gv]

    # Allow for the possibility that no analog (or group, in the new code) is retrieved.
    p_retrieve = [i/(pi_nr + sum(s_x_gv)) for i in s_x_gv]

    p_cum = [sum(p_retrieve[0:i+1]) for i in xrange(len(p_retrieve))]

    for i in xrange(len(groups)):
        print '%s Similarity: %0.2f P Retrive: %0.2f' % (groups[i].name, similarity[i], p_retrieve[i])

    sample = random.random()
    print 'Sample:', sample
    if sample > p_cum[-1]:
        group_index = None
    elif sample <= p_cum[0]:
        group_index = 0
    else:
        for i in xrange(len(p_cum)):
            if p_cum[i] < sample <= p_cum[i+1]:
                group_index = i+1

    if group_index is None:
        print 'No group retrieved.  source:', __file__
        seq = None
    else:
        group = groups[group_index]
        seq = group.sequence
        if seq.retrieved == True:
            seq = None
        else:
            seq.retrieved = True        

    # Reset prop.max_act, group.weber_fraction
    for prop in analogs['ltm'].props:
        prop.max_act = 0.0
        prop.old_max_act = 0.0
    for group in analogs['ltm'].groups:
        group.similarity_strength = 0.0
        group.act_memory = []

    return seq

def move_sequence(analogs, from_analog, to_analog, sequence):
    seq_i = analogs[from_analog].sequences.index(sequence)
    analogs[to_analog].sequences.append(analogs[from_analog].sequences.pop(seq_i))
    for group in sequence.ordered_groups:
        g_i = analogs[from_analog].groups.index(group)
        analogs[to_analog].groups.append(analogs[from_analog].groups.pop(g_i))
    for prop in sequence.props:
        p_i = analogs[from_analog].props.index(prop)
        analogs[to_analog].props.append(analogs[from_analog].props.pop(p_i))
    for rf in sequence.rfbindings:
        rf_i = analogs[from_analog].rfbindings.index(rf)
        analogs[to_analog].rfbindings.append(analogs[from_analog].rfbindings.pop(rf_i))
    for pred in sequence.preds:
        pred_i = analogs[from_analog].preds.index(pred)
        analogs[to_analog].preds.append(analogs[from_analog].preds.pop(pred_i))
    for obj in sequence.objs:
        o_i = analogs[from_analog].objs.index(obj)
        analogs[to_analog].objs.append(analogs[from_analog].objs.pop(o_i))

