## persephone.py    Derek Devnich    14 March 2007

'''Welcome to Persephone, a model of explanation by analogy.

 * This code runs in Python 2.4 and 2.5, and may run in earlier versions of Python.

 * If you want the GUI, you need to install PyGame.  PyGame is available at www.pygame.org


* * * Warnings for Maintainers and Confession of Sins * * *

  1. Persephone pretends to be object-oriented, but does not properly honor
  encapsulation.  Many of the functions in network.py and update.py directly
  change the attributes of instances, rather than calling getter and setter
  methods belonging to those instances.  As a result, a change to the code
  may actually require multiple parallel changes.
  
  2. Persphone has limited exception handling.  In particular, the methods in
  symparser.py that create the network from the simulation description are not
  forgiving of syntax errors.
  
  3. At start-up, the model creates an instance of the Symparser class.  The
  Symparser class follows the "Singleton" pattern: It's a glorified name-space
  with no good reason to live.  The model expects that there will be one, and
  only one, Symparser instance; creating multiple instances will probably result
  in bad behavior.

The following simulation files come with this distribution:
  coffee
  coffee2
  coffee_nested
  coffee2_nested
  
Y'all have fun now!'''

if __name__ == '__main__':
    print __doc__
    
#-------------------------------------------------
# Setup
#-------------------------------------------------

# System imports
import sys, os, random, copy
random.seed(os.urandom(99))
sys.setcheckinterval(10000)             # Default is 100

# Model imports
import display
import symparser
import perstypes
import network as nw
import update as up
import memory as mem
#import lisaout
from persparms import *

sp = symparser.Symparser()

# Model control variables
model = ['shell','run']             # command line/simulation run switch
sym_file = None                     # Symfile name
gui = False                         # gui switch
pause = True                        # pause switch
fps = 30                            # frames per second
frame_ratio = 1.5                   # frame rate multiplier
simulation_done = False

# Drawing containers
update_list = []
screen = None

#-------------------------------------------------
# Start the interactive shell
#-------------------------------------------------
while model[0] == 'shell':
    
    if gui == True:
        gui_message = 'ON.  Press |G| to turn them off.'
    elif gui == False:
        gui_message = 'OFF.  Press |G| to turn them on.'
    
    print ''
    print ' ----------------------------------'
    print '  * * * Persephone Main Menu * * *'
    print ' ----------------------------------'
    print ''
    print ' Please choose from the following options:'
    print '  |F|ile:  Choose a simulation file to run'
    print '  |C|lear: Current simulation file is "', str(sym_file), '".  Press |C| to clear.'
    print '  |R|un:   Run the simulation'
    print '  |G|ui:   Graphics are', gui_message
    print '  |A|:     Run the default selection'
    print '  |E|xit:  Exit the model shell'
    print ''
    command = raw_input(' PERSEPHONE=> ')

    if command.lower() == 'a':
        sym_file = 'coffee2_nested'
        print 'You are about to run', sym_file
        import coffee2_nested
        gui = False
    
    elif command.lower() == 'f':
        if sym_file is not None:
            print '\n  * * A simulation has already been selected.  Press |C| to clear.'
        else:
            sym_file = raw_input('\n  * * Please enter a sym file name: ')
            try:
                exec('import ' + sym_file)
            except ImportError:
                print '\n  > > "', sym_file, '" does not exist.  Please try again.'
                sym_file = None

    elif command.lower() == 'c':
        sym_file = None
        
    elif command.lower() == 'r':

        if sym_file is not None:
            model.append(model.pop(0))
            # Create network from symfile
            sp.sym = eval(sym_file).Symfile()
            sp.parse_symfile()           
            simulation_sequence = zip(sp.sym.action_flags, sp.sym.driver_types)
        else:
              print '\n  > > You need to enter a simulation file name.'
        
    elif command.lower() == 'g':
        if gui == True:
            gui = False
        elif gui == False:
            gui = True
    elif command.lower() == 'e':
        sys.exit(0)
    else:
        print '\n  > > That\'s not an option.  Please try again.'

#-------------------------------------------------
# Start the simulation
#-------------------------------------------------
while model[0] == 'run' and sp.sym.max_cycles > 0:
    sp.sym.max_cycles -= 1

    #==> TO DO: Re-parse target for multiple runs

    # Instantiate pygame stuff here
    if gui == True:
        try:
            import pygame
            from pygame.locals import *
            network_display = display.NetworkDisplay()
            clock = pygame.time.Clock()
            screen = pygame.display.set_mode((1024,768))#((1280,1024))#,FULLSCREEN)
        except ImportError:
            print 'PyGame not installed.  No GUI for you!'
            gui = False
            
    if gui == True:

        # Build initial display
        network_display.purge_display(screen)
        for analog in sp.analogs.itervalues():
            if analog.name in network_display.nodes:
                network_display.add_label(screen, analog.name, update_list, [analog.name, 'label'])
                network_display.add_analog(analog, screen, update_list)
        network_display.add_semantics(sp.pred_semantics, sp.obj_semantics, sp.episode_semantics,
                                      sp.link_semantics, screen, update_list)
        
        # Draw initial network state to screen
        network_display.add_status(screen, network_display.input[0], kwd=['status','input'])
        network_display.add_status(screen, str(fps), kwd=['status','speed'])
        network_display.add_status(screen, str(pause), kwd=['status','pause'])
        # Draw the network
        network_display.draw_network(screen, update_list)

    retrieved = True
    while retrieved == True:

        for sequence in sp.analogs['target'].sequences:
            sp.set_mappable(sp.sym.direction, sequence, driver=None, action_flag='retrieve',
                            reset_only=False, to_analog='source')      
        sp.clear_hebbs()
        
        sim_flags = copy.copy(simulation_sequence)
        while any(sim_flags):
            
            sim_flag = sim_flags[0]
            print ' * * Current simulation task => ', sim_flag
       
            #------------------------------------------------------------
            # Set driver and recipients; initialize everything
            #------------------------------------------------------------
            driver = None
            for analog in sp.analogs.itervalues():
                analog.analog_type = None
                analog.top_down_ok = False
                analog.ready_to_learn = False

                if analog.name == sim_flag[1]:
                    analog.analog_type = 'driver'
                    driver = analog.name
                else:
                    if analog.name == 'ltm':
                        if sim_flag[0] == 'retrieve':
                            analog.analog_type = 'recipient'
                    elif analog.name == 'target':
                        if sim_flag[0] == 'infer' or sim_flag[0] == 'induce':
                            analog.analog_type = 'recipient'
                    elif analog.name == 'source':
                        if sim_flag[0] == 'map':
                            analog.analog_type = 'recipient'
                    elif analog.name == 'schema':
                        if sim_flag[0] == 'induce':
                            analog.analog_type = 'recipient'

            # Depending on driver and action_flag, set target or source as mappable
            if sim_flag[0] == 'map' and driver == 'target':
                for sequence in sp.analogs['source'].sequences:
                    sp.set_mappable(sp.sym.direction, sequence, driver, sim_flag[0],
                                    reset_only=False, to_analog='target')
            elif sim_flag[0] == 'infer' and driver == 'source':
                for sequence in sp.analogs['source'].sequences:
                    sp.set_mappable(sp.sym.direction, sequence, reset_only=True)
                for sequence in sp.analogs['target'].sequences:
                    sp.set_mappable(sp.sym.direction, sequence, driver, sim_flag[0],
                                    reset_only=False, to_analog='source')

            # Initialize retrieval            
            if sim_flag[0] == 'retrieve':
                for prop in sp.analogs['ltm'].props:
                    prop.max_act = 0.0
                    prop.old_max_act = 0.0
                for group in sp.analogs['ltm'].groups:
                    group.similarity_strength = 0.0

            if sim_flag[0] == 'induce' and sim_flag[1] == 'source' and gui == True:
                network_display.clear_analog('schema')

            # Indicate driver in GUI
            if gui == True:
                for analog_name, contents in network_display.nodes.iteritems():
                    if analog_name != 'status':
                        for node in network_display.nodes[analog_name]['label']:
                            if analog_name == driver:
                                node.update_label(update_list, screen, label_type='driver')
                            else:
                                node.update_label(update_list, screen, label_type='normal')
                if sim_flag[0] == 'retrieve':
                    # Draw these units
                    network_display.clear_analog('ltm')
                    network_display.clear_analog('source')
                    network_display.purge_display(screen)
                    network_display.add_label(screen, 'target', update_list, ['target', 'label'])
                    network_display.add_analog(sp.analogs['target'], screen, update_list)
                    network_display.add_label(screen, 'ltm', update_list, ['ltm', 'label'])
                    network_display.add_analog(sp.analogs['ltm'], screen, update_list)
                    network_display.add_semantics(sp.pred_semantics, sp.obj_semantics, sp.episode_semantics,
                                                  sp.link_semantics, screen, update_list)
                    network_display.draw_network(screen, update_list)
  
            #---------------------------------------------------
            # Set mode of driver props and set firing sequence
            #---------------------------------------------------
            '''The current phase set can contain one or more driver props.'''

            # Order only at the level of props
            for sequence in sp.analogs[driver].sequences:
                sp.set_group_firing_sequence(sequence, sp.sym.direction, sim_flag[0])
            #print 'Group queue:', [i.name for i in sp.group_queue if i is not None]

            ### This is probably redundant...                
            for analog in sp.analogs.itervalues():
                analog.update_global_inhibitor(flag='major')
                       
            while len(sp.group_queue) > 0:
                
                firing_episode = sp.group_queue.pop(0)
                if firing_episode is not None:
                    firing_episode.set_fired(True)

                # Do selection of probes here if action_flag == 'retrieve'
                sp.set_prop_firing_sequence(driver, firing_episode, sim_flag[0])

                if sim_flag[0] == 'retrieve':
                    print 'Prop queue:', [i.name for i in sp.prop_queue]

                while len(sp.prop_queue) > 0:
                    
                    # This yields a prop or a list of props
                    phase_set = sp.prop_queue.pop(0)
                    
                    #-------------------------------------------------------
                    # Run the phase set for the current set of driver props
                    #-------------------------------------------------------

                    up.set_phase_set(phase_set, firing_episode)
                    
                    # Start phase set with top_down_ok == False
                    up.set_top_down_ok(sp.analogs, top_down_flag=False)

                    # This should work...top_down starts out ok for schema induction
                    if sim_flag[0] == 'induce':
                        for analog in sp.analogs.itervalues():
                            if analog.analog_type == 'recipient':
                                analog.top_down_ok = True
                                analog.ready_to_learn = True
                    
                    # Set number of cycles
                    if type(phase_set) != list:
                        phase_set = [phase_set]
                    phase_cycle = len(phase_set)
                    props_fired = 1
                    if sim_flag[0] == 'retrieve':
                        phase_set = phase_set * rho_mem
                    else:
                        phase_set = phase_set * rho_cy
                        
                    for prop in phase_set:

                        # When starting new Prop, update global inhibitor for all driver units
                        sp.analogs[driver].update_global_inhibitor(flag='major')

                        #----------------------------------------------------
                        # Fire each of the current Prop's RFs once
                        #----------------------------------------------------
                        
                        for rf in prop.rf:

                            #-------------------------------
                            # Begin current RF
                            #-------------------------------

                            # Initialize all props to 'neutral'
                            up.initialize_props(sp.analogs)

                            # Set parent and child mode on a prop-by-prop basis
                            rf.prop.mode = 'parent'
                            if type(rf.obj) is perstypes.Prop:
                                rf.obj.mode = 'child'

                            # Update the global inhibitor in driver RFs and OPs
                            sp.analogs[driver].update_global_inhibitor(flag='minor')

                            # Update global inhibitor in all recipient units
                            for analog in sp.analogs.itervalues():
                                if analog.analog_type == 'recipient':
                                    analog.update_global_inhibitor(flag='major')

                            up.update_semantic_inhibitor(sp.pred_semantics)
                            up.update_semantic_inhibitor(sp.obj_semantics)
                            up.update_semantic_inhibitor(sp.episode_semantics)
                            up.update_semantic_inhibitor(sp.link_semantics)
                            
                            # Set RF activation in driver
                            rf.net_input = 1.0
                            rf.act = 1.0
                            
                            for j in xrange(max_t):

                                #----------------------------------------------------
                                # GUI controls
                                #----------------------------------------------------

                                if gui == True:
                                    clock.tick(fps)

                                # While paused, check for un-pause and other keystrokes
                                if gui == True:
                                    while pause == True:
                                        wait = pygame.event.wait()
                                        if not hasattr(wait,'key'):
                                            continue
                                        # Check for un-pause
                                        if wait.key == K_SPACE and wait.type == KEYDOWN:
                                            pause = False
                                        # Exit
                                        elif wait.key == K_ESCAPE:
                                            sys.exit(0)
                             
                                # Get user input
                                if gui == True:
                                    for event in pygame.event.get():
                                        # Pass on non-key events
                                        if not hasattr(event,'key'):
                                            continue
                                        # Increase frame rate
                                        if event.key == K_RIGHT and event.type == KEYDOWN and gui == True:
                                            fps = fps * frame_ratio
                                        # Decrease frame rate
                                        elif event.key == K_LEFT and event.type == KEYDOWN and gui == True:
                                            fps = fps/frame_ratio
                                        # Pause
                                        elif event.key == K_SPACE and event.type == KEYDOWN and gui == True:
                                            if pause == False:
                                                pause = True
                                        # Exit
                                        elif event.key == K_ESCAPE:
                                            sys.exit(0)

                                #----------------------------------------------------
                                # Innermost loop of mapping code
                                #----------------------------------------------------

                                # Exit mapping/inference loop when activations in recipients settle
                                if sim_flag[0] != 'retrieve': 
                                    if j > min_t:
                                        exit_flag = True
                                        for analog in sp.analogs.itervalues():
                                            if analog.analog_type == 'recipient':
                                                pred_delta_count = sum(1 for i in analog.preds if abs(i.act - i.old_act) > theta_h)
                                                obj_delta_count = sum(1 for i in analog.objs if abs(i.act - i.old_act) > theta_h)
                                                rf_delta_count = sum(1 for i in analog.rfbindings if abs(i.act - i.old_act) > theta_h)
                                                prop_delta_count = sum(1 for i in analog.props if abs(i.act - i.old_act) > theta_h)
                                                group_delta_count = sum(1 for i in analog.groups if abs(i.act - i.old_act) > theta_h)
                                                if pred_delta_count > 0 or obj_delta_count > 0 or rf_delta_count > 0 or prop_delta_count > 0 or group_delta_count > 0:
                                                    exit_flag = False
                                        if exit_flag == True:
                                            #pass
                                            #print j, 'iterations'
                                            break
                                        
                                # Exit retrieval loop when Prop max_acts in LTM settle
                                if sim_flag[0] == 'retrieve':
                                    if j > min_t:
                                        exit_flag = True
                                        for analog in sp.analogs.itervalues():
                                            if analog.analog_type == 'recipient' and analog.name == 'ltm':
                                                max_act_delta_count = sum(1 for i in analog.props if abs(i.max_act - i.old_max_act) > theta_h)
                                                group_delta_count = sum(1 for i in analog.groups if abs(i.act - i.old_act) > theta_h)
                                                if max_act_delta_count > 0 or group_delta_count > 0:
                                                    exit_flag = False
                                        if exit_flag == True:
                                            #print j, 'iterations.'
                                            break
                                            #pass
                                                
                                # Update the modes of all Props in the recipient
                                up.set_recipient_mode(sp.analogs, driver, sp.hebbs, phi,j)

                                # Get active driver units' activations and max hebb weights
                                if sim_flag[0] == 'map' or sim_flag[0] == 'infer' or sim_flag[0] == 'induce' or sim_flag[0] == 'perceive' or sim_flag[0] == 'world_infer' or sim_flag[0] == 'test' or sim_flag[0] == 'evaluate':
                                    up.get_driver_max_hebbs(sp.analogs, driver, sp.hebbs, theta_c, j)
                                
                                # Update inputs to driver units in the phase set
                                up.traverse_tokens(sp.analogs, driver, sp.hebbs, hebb_bias, j, sim_flag[0],
                                                   analog_flag='driver')

                                up.traverse_semantics(sp.analogs, sp.pred_semantics)
                                up.traverse_semantics(sp.analogs, sp.obj_semantics)
                                up.traverse_semantics(sp.analogs, sp.episode_semantics)
                                up.traverse_semantics(sp.analogs, sp.link_semantics)

                                # Update inputs to all recipient units
                                up.traverse_tokens(sp.analogs, driver, sp.hebbs, hebb_bias, j, sim_flag[0],
                                                   analog_flag='recipient')

                                # Update activations of all units
                                up.update_activations(sp.analogs, gamma, delta, sim_flag[0])
                                if len(sp.hebbs) > 0:
                                    up.get_hebb_buffers(sp.hebbs)

                                # During retrieval, update maximum activation value for all props in LTM
                                if sim_flag[0] == 'retrieve':
                                    mem.update_prop_max_act(sp.analogs)
                                    mem.update_group_act_memory(sp.analogs)
                                    #print [(i.name, i.max_act) for i in sp.analogs['ltm'].props if i.act > 0.1]

                                # Check whether to license self-supervised learning
                                if sim_flag[0] == 'infer' or sim_flag[0] == 'induce' or sim_flag[0] == 'world_infer':
                                    for analog in sp.analogs.itervalues():
                                        if analog.analog_type == 'recipient' and analog.ready_to_learn == False:
                                            nw.get_importance(analog)
                                            nw.license_learning(analog, pi_m)

                                # During mapping, create hebbs
                                if sim_flag[0] == 'map' or sim_flag[0] == 'infer' or sim_flag[0] == 'induce' or sim_flag[0] == 'perceive' or sim_flag[0] == 'world_infer' or sim_flag[0] == 'test' or sim_flag[0] == 'evaluate':
                                        
                                    # "Retrieve" recipient units and create hebbs
                                    nw.create_hebbs(sp.analogs, driver, sp.hebbs, theta_r)

                                    # If the model is ready to learn, run these four methods:
                                    # (requires analog.ready_to_learn == True and analog.top_down_ok == True)
                                    # NB: After a unit is inferred, you must update everyone's activations before
                                    # connecting it.
                                    nw.connect_new_units(sp.analogs, sp.sym.direction, driver,
                                                         sp.pred_semantics, sp.obj_semantics, sp.episode_semantics,
                                                         sp.link_semantics, theta_c)
                                    nw.infer_new_units(sp.analogs, sp.sym.direction, driver, sp.hebbs,
                                                       network_display, screen, update_list, pi_i,j)
                                    nw.update_pred_semantic_weights(sp.analogs, mu)
                                    nw.update_obj_semantic_weights(sp.analogs, mu)
                                    nw.update_group_semantic_weights(sp.analogs, mu)

                                #----------------------------------------------------
                                # GUI update for current iteration
                                #----------------------------------------------------
                                if gui == True:
                                    for node_type, node_list in network_display.nodes['target'].iteritems():
                                        if node_type != 'label':
                                            for node in node_list:
                                                node.update_mode(network_display)
                                                node.draw_border(screen, update_list, retrieve_flag=True)
                                                node.update_act(screen, update_list)
                                    for node_type, node_list in network_display.nodes['source'].iteritems():
                                        if node_type != 'label':
                                            for node in node_list:
                                                node.update_mode(network_display)
                                                node.draw_border(screen, update_list, retrieve_flag=True)
                                                node.update_act(screen, update_list)
                                    for node_type, node_list in network_display.nodes['ltm'].iteritems():
                                        if node_type != 'label':
                                            for node in node_list:
                                                node.update_mode(network_display)
                                                node.draw_border(screen, update_list, retrieve_flag=True)
                                                node.update_act(screen, update_list)
                                    for node_type, node_list in network_display.nodes['semantic'].iteritems():
                                        if node_type != 'label':
                                            for node in node_list:
                                                node.update_act(screen, update_list)

                                    # Update everything on this loop
                                    pygame.display.update(update_list)
                                    update_list = []
                                    
                            #--------------------------------
                            # End of current RF
                            #--------------------------------

                        props_fired += 1

                        # After first time through, top_down_ok == True
                        if sim_flag[0] == 'map' or sim_flag[0] == 'infer' or sim_flag[0] == 'induce' or sim_flag[0] == 'perceive' or sim_flag[0] == 'world_infer' or sim_flag[0] == 'test' or sim_flag[0] == 'evaluate':
                            if props_fired > phase_cycle:
                                for analog in sp.analogs.itervalues():
                                    if analog.analog_type == 'recipient':
                                        analog.top_down_ok = True

                    #--------------------------------
                    # End of phase set
                    #--------------------------------

                    if sim_flag[0] == 'map' or sim_flag[0] == 'infer' or sim_flag[0] == 'induce' or sim_flag[0] == 'perceive' or sim_flag[0] == 'world_infer' or sim_flag[0] == 'test' or sim_flag[0] == 'evaluate':

                        # Update mapping connections
                        nw.find_consistent_hebbs(sp.analogs, driver, sp.hebbs)                
                        up.get_hebb_div_max(sp.hebbs)
                        up.apply_hebb_div_max(sp.hebbs)

                        # Old method for implementing competition between hebbs
                        #up.get_hebb_sub_max(analogs, hebbs)

                        # New method for implementing competition between hebbs
                        up.settle_hebbs(sp.hebbs, rho_he, rho_hi, gamma_h, delta_h, theta_h)
                        up.update_hebb_weights(sp.hebbs, eta)
                        nw.kill_hebbs(sp.analogs, sp.hebbs, theta_k)
                        up.flush_hebbs(sp.hebbs)
          
                        #lisaout.write_mappings(analogs, driver, hebbs, fid)

                    # Update global inhibitor for entire recipient
                    for analog in sp.analogs.itervalues():
                        if analog.analog_type == 'recipient':
                            analog.update_global_inhibitor(flag='major')

                    # Reset all phase set flags to False
                    for group in sp.analogs[driver].groups:
                        group.phase_set_flag = False
                    for prop in sp.analogs[driver].props:
                        prop.phase_set_flag = False
                    for rf in sp.analogs[driver].rfbindings:
                        rf.phase_set_flag = False
                    for pred in sp.analogs[driver].preds:
                        pred.phase_set_flag = False
                    for obj in sp.analogs[driver].objs:
                        obj.phase_set_flag = False

            # Print statements for end of action
            sp.print_analog(driver, group_flag=True, prop_flag=True, concise_groups=False)
            for analog in sp.analogs.itervalues():
                if analog.analog_type == 'recipient':
                    sp.print_analog(analog.name, group_flag=True, prop_flag=True, concise_groups=False)
            sp.print_hebbs()
##            check_mappings = nw.find_consistent_hebbs(sp.analogs, driver, sp.hebbs, eval_flag=True)
##            print 'Mappings:', check_mappings
            
            #--------------------------------
            # Retrieve after all props fire
            #--------------------------------

            if sim_flag[0] == 'retrieve':
                    
                # Do calculations and retrieve source
                mem.get_retrieval_fit(sp.analogs, sp.sym.direction)
                for seq in sp.analogs['source'].sequences:
                    for group in seq.ordered_groups:
                        group.set_fired(False)
                    mem.move_sequence(sp.analogs, 'source', 'ltm', seq)
                seq = mem.retrieve_analog(sp.analogs, driver, pi_nr)
                
                if seq is not None:
                    mem.move_sequence(sp.analogs, 'ltm', 'source', seq)
                    sp.set_mappable(sp.sym.direction, seq, to_analog=None, reset_only=True)
                    for group in seq.ordered_groups:
                        group.set_fired(False)
                else:
                    retrieved = False
                    sim_flags = []
                    
                    # Save the target to LTM if anything was inferred
                    for sequence in sp.analogs['target'].sequences:
                        inferred_units = False
                        if any(i for i in sequence.ordered_groups if i.inferred == True):
                            inferred_units = True
                        elif any(i for i in sequence.props if i.inferred == True):
                            inferred_units = True
                        elif any(i for i in sequence.rfbindings if i.inferred == True):
                            inferred_units = True
                        elif any(i for i in sequence.preds if i.inferred == True):
                            inferred_units = True
                        elif any(i for i in sequence.objs if i.inferred == True):
                            inferred_units = True
                        if inferred_units == True:
                            sp.set_mappable(sp.sym.direction, sequence, to_analog=None, reset_only=True)
                            mem.move_sequence(sp.analogs, 'target', 'ltm', sequence)
                    break
                    break
                    
                if retrieved == True and gui == True:
                    # Draw these units
                    network_display.clear_analog('ltm')
                    network_display.clear_analog('source')
                    network_display.purge_display(screen)
                    network_display.add_label(screen, 'target', update_list, ['target', 'label'])
                    network_display.add_analog(sp.analogs['target'], screen, update_list)
                    network_display.add_label(screen, 'source', update_list, ['source', 'label'])
                    network_display.add_analog(sp.analogs['source'], screen, update_list)
                    network_display.add_semantics(sp.pred_semantics, sp.obj_semantics, sp.episode_semantics,
                                                  sp.link_semantics, screen, update_list)
                    network_display.draw_network(screen, update_list)

            #---------------------------
            # Cycle sim_flags
            #---------------------------
            if len(sim_flags) > 0:
                sim_flags.pop(0)
                

    if sp.sym.max_cycles == 0:
        model.append(model.pop(0))
