'''persparms.py             25 February 2006

These are GLOBAL parameters for Persephone.  Do not change during a model run.'''

phi = 0.01      # threshold for changing prop mode from 'neutral' [Prop.setmode()]
gamma = 0.3     # growth rate for Token activation leaky integrator [Token.get_delta()].
                # 0.3 and 0.4 both work well; 0.4 appears to be slightly faster.
delta = 0.1     # decay rate for Token activation leaky integrator.  0.1 and 0.05 both work well;
                # 0.05 appears to be slightly faster.
theta_i = 0.2   # activation threshold for using Token units to normalize activation
eta = 0.9       # growth rate for hebb weights
theta_r = 0.5   # threshold for retrieving a recipient unit into WM [create_hebbs()]

## This is a hacked threshold for retrieving objects.  The OP.get_bu_input() normalization requires
## OPs to be at least 51% similar to one another at maximum activation in order to retrieved.
## As a matter of practice, objects are usually less similar to one another than predicates.
## This parameter is left over from an earlier version of the code, and may be unnecessary now.
## theta_r_hack = 0.5

theta_k = 0.1   # threshold for killing off hebbs at the end of a phase set
theta_m = 0.9   # threshold for saying two units have mapped
theta_c = 0.1   # threshold for connecting inferred units
pi_nr = 0.1     # chance of not retrieving any LTM item; an item's chance of retrieval scales between 0 and 1.
theta_f = 0.5   # threshold for saying that the constituents of an LTM group "fit" the retrieval probe.
mu = 0.2        # learning rate for connecting semantics to inferred units
pi_m = 0.7      # weighted proportion of reciepient mapped to license self-supervised learning

pi_i = 2.0      # controls proportion of inhibited units required to license inference
                # The original value is 2; however, any value larger than 1 causes isomorphic
                # analogs to do infererence, b/c the last units are still unmapped when the
                # number of inhibited units passes the threshold for inference
                
rho_rf_i = 3    # inhibition multiplier for RFs.  This must be > 2.  3 seems to work well.
                # 4 also works, but seems to take longer.
rho_g_i = 3     # inhibition multiplier that groups use on RFs not their own
rho_g_p = 2     # inhibition multiplier that groups use on props not their own
rho_g = 2       # inhibition multiplier that groups use to inhibit each other
#rho_op_i = 1    # inhibition multiplier for OPs and child props.  1 works well, so this parameter
                # isn't actually used in the code.  Depending on what else is going on, 2 has also
                # been seen to work.
rho_cy = 2      # number of cycles for RFs in phase set during mapping.  2 works for serial mapping,
                # but batched mapping requires the chance to go through units after they've been
                # hebbed.
rho_mem = 1     # number of cycles for RFs in phase set during retrieval
max_t = 75      # maximum number of time steps per RF firing
min_t = 10      # minimum number of time steps per RF firing

gamma_h = 0.2   # Growth rate for hebb activations
delta_h = 0.05  # Decay rate for hebb activations
theta_h = 0.01  # Settling threshold for hebb activations
rho_hi = 20     # Multiplier for competing hebb inhibition in hebb weight learning
rho_he = 3      # Multiplier for consistent hebb excitation in hebb weight learning
hebb_bias = 2   # Multiplier used to enhance the effects of hebb input (excitatory and inhibitory)
                # for Token.get_total_hebb_input().  This allows mapped units to strongly constrain
                # who gets to join up with newly-inferred units in, e.g., Towers of Hanoi.
##pred_influence = 0.3    # multiplier for pred.bu_input during evaluation.  Attenuating pred input
##                        # gives literal object-to-object mappings from target effect to world,
##                        # increasing the chance of finding matches out of place (and thereby
##                        # detecting wrong inconsistent effects)

gui_x = 90      # width of unit in GUI in pixels
gui_y = 40      # height of unit
label_y = 14    # height of analog label

e_lambda = 0.5  # Exponential distribution parameter used for importance-based firing.  Lower numbers
                # yield a flatter density function.
beta = 0.2      # Learning rate for reinforcement learning

## k = 1        # normalization constant; it's hard-coded as 1 in every function that has one






